@@ -1,32 +1,56 @@
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at https://mozilla.org/MPL/2.0/.
-import pandas as pd
+from __future__ import annotations
+
 import numpy as np
+import pandas as pd
+
 from chainladder.utils.utility_functions import num_to_nan
+from typing import TYPE_CHECKING
+
+
+try:
+    import dask.bag as db
+except ImportError:
+    db = None
+
+if TYPE_CHECKING:
+    from chainladder import Triangle
+    from collections.abc import Callable
+    from numpy.typing import ArrayLike
+    from types import ModuleType
+    from typing import Type
+
 
 
 class TriangleGroupBy:
-    def __init__(self, obj, by, axis=0):
+    def __init__(self, obj, by, axis=0, **kwargs):
         self.obj = obj.copy()
-        axis = self.obj._get_axis(axis)
-        self.by = by
-        if axis > 1:
-            raise ValueError(
-                "Use grain method to group by origin and development axes."
-            )
         self.axis = self.obj._get_axis(axis)
-        if self.axis == 0:
+        self.by = by
+        if kwargs.get("groups", None):
+            self.groups = kwargs.get("groups", None)
+        elif self.axis == 0:
             self.groups = obj.index.groupby(by)
-        if self.axis == 1:
+        elif axis == 1:
             self.groups = pd.DataFrame(obj.columns).groupby(by)
+        else:
+            self.groups = pd.DataFrame(by).groupby(by)
+
+    def __getitem__(self, key):
+        return TriangleGroupBy(
+            obj=self.obj.__getitem__(key),
+            by=self.by,
+            axis=self.axis,
+            groups=self.groups,
+        )
 
 
 class TrianglePandas:
-    def to_frame(self, origin_as_datetime=False, keepdims=False, implicit_axis=False,
-                 *args, **kwargs):
+    def to_frame(self, origin_as_datetime=True, keepdims=False,
+                 implicit_axis=False, *args, **kwargs):
         """ Converts a triangle to a pandas.DataFrame.
-
         Parameters
         ----------
         origin_as_datetime : bool
@@ -44,12 +68,17 @@ class TrianglePandas:
             pandas.DataFrame representation of the Triangle.
         """
         axes = [num for num, item in enumerate(self.shape) if item > 1]
+
         if keepdims:
             is_val_tri = self.is_val_tri
             obj = self.val_to_dev().set_backend("sparse")
             out = pd.DataFrame(obj.index.iloc[obj.values.coords[0]])
             out["columns"] = obj.columns[obj.values.coords[1]]
-            out["origin"] = obj.odims[obj.values.coords[2]]
+            missing_cols = list(set(self.columns) - set(out['columns']))
+            if origin_as_datetime:
+                out["origin"] = obj.odims[obj.values.coords[2]]
+            else:
+                out["origin"] = obj.origin[obj.values.coords[2]]
             out["development"] = obj.ddims[obj.values.coords[3]]
             out["values"] = obj.values.data
             out = pd.pivot_table(
@@ -62,70 +91,115 @@ class TrianglePandas:
 
             valuation = pd.DataFrame(
                 obj.valuation.values.reshape(obj.shape[-2:], order='F'),
-                index=obj.odims, columns=obj.ddims
+                index=obj.odims if origin_as_datetime else obj.origin, 
+                columns=obj.ddims
             ).unstack().rename('valuation').reset_index().rename(
                 columns={'level_0': 'development', 'level_1': 'origin'})
 
             val_dict = dict(zip(list(zip(
                 valuation['origin'], valuation['development'])),
                 valuation['valuation']))
-            out['valuation'] = out.apply(
-                lambda x: val_dict[(x['origin'], x['development'])], axis=1)
-            col_order = list(out.columns)
+            if len(out) > 0:
+                out['valuation'] = out.apply(
+                    lambda x: val_dict[(x['origin'], x['development'])], axis=1)
+            else:
+                out['valuation'] = self.valuation_date
+            col_order = list(self.columns)
             if implicit_axis:
-                col_order = ['origin', 'development', 'valuation'] + col_order[2:-1]
+                col_order = ['origin', 'development', 'valuation'] + col_order
             else:
                 if is_val_tri:
-                    col_order = ['origin', 'valuation'] + col_order[2:-1]
+                    col_order = ['origin', 'valuation'] + col_order
                 else:
-                    col_order = ['origin', 'development'] + col_order[2:-1]
+                    col_order = ['origin', 'development'] + col_order
+            for col in set(missing_cols) - self.virtual_columns.columns.keys():
+                out[col] = np.nan
+            for col in set(missing_cols).intersection(self.virtual_columns.columns.keys()):
+                out[col] = out.fillna(0).apply(self.virtual_columns.columns[col], 1)
+                out.loc[out[col] == 0, col] = np.nan
+
             return out[col_order]
-        if self.shape[:2] == (1, 1):
-            return self._repr_format(origin_as_datetime)
-        elif len(axes) in [1, 2]:
-            tri = np.squeeze(self.set_backend("numpy").values)
-            axes_lookup = {
-                0: self.kdims,
-                1: self.vdims,
-                2: self.origin,
-                3: self.development,
-            }
-            if axes[0] == 0:
-                idx = self.index.set_index(self.key_labels).index
-            else:
-                idx = axes_lookup[axes[0]]
-            if len(axes) == 2:
-                return pd.DataFrame(
-                    tri, index=idx, columns=axes_lookup[axes[1]]
-                ).fillna(0)
-            if len(axes) == 1:
-                return pd.Series(tri, index=idx).fillna(0)
+
+        # keepdims = False
         else:
-            return self.to_frame(
-                origin_as_datetime=origin_as_datetime, keepdims=True,
-                implicit_axis=implicit_axis)
+            if self.shape[:2] == (1, 1):
+                return self._repr_format(origin_as_datetime)
+
+            elif len(axes) in [1, 2]:
+                tri = np.squeeze(self.set_backend("numpy").values)
+                axes_lookup = {
+                    0: self.kdims,
+                    1: self.vdims,
+                    2: self.origin,
+                    3: self.development,
+                }
+
+                if axes[0] == 0:
+                    idx = self.index.set_index(self.key_labels).index
+                else:
+                    idx = axes_lookup[axes[0]]
+
+                if len(axes) == 1:
+                    return pd.Series(tri, index=idx).fillna(0)
+
+                elif len(axes) == 2:
+                    return pd.DataFrame(
+                        tri, index=idx, columns=axes_lookup[axes[1]]
+                    ).fillna(0)
+
+            else:
+                return self.to_frame(
+                    origin_as_datetime=origin_as_datetime, keepdims=True,
+                    implicit_axis=implicit_axis)
 
     def plot(self, *args, **kwargs):
-        """ Passthrough of pandas functionality """
-        return self.to_frame().plot(*args, **kwargs)
+        """Passthrough of pandas functionality"""
+        return self.to_frame(origin_as_datetime=False).plot(*args, **kwargs)
 
     def hvplot(self, *args, **kwargs):
-        """ Passthrough of pandas functionality """
-        df = self.to_frame()
+        """Passthrough of pandas functionality"""
+        df = self.to_frame(origin_as_datetime=True)
         if type(df.index) == pd.PeriodIndex and len(df.columns) > 1:
             df.index = df.index.to_timestamp(how="s")
         return df.hvplot(*args, **kwargs)
 
-    def _get_axis(self, axis):
+    @staticmethod
+    def _get_axis(axis: str | int| None) -> int:
+        """
+        Returns the integer representation of the requested axis.
+
+        Parameters
+        ----------
+        axis: str | int | None
+            String or integer representation of the requested axis. If
+            supplied as a string, returns the integer representation. If
+            supplied as an integer, returns the same integer.
+
+        Returns`
+        -------
+        The integer representation of the requested axis
+
+        """
+
         ax = {
             **{0: 0, 1: 1, 2: 2, 3: 3},
             **{-1: 3, -2: 2, -3: 1, -4: 0},
             **{"index": 0, "columns": 1, "origin": 2, "development": 3},
         }
-        return ax.get(axis, 0)
+
+        try:
+            return ax[axis]
+        except KeyError:
+            if axis is None:
+                return 0
+            else:
+                raise ValueError(
+                    "Invalid axis specified. Please specify the correct string or "
+                    "integer representation of the desired axis."
+                )
 
     def dropna(self):
-        """  Method that removes orgin/development vectors from edge of a
+        """Method that removes origin/development vectors from edge of a
         triangle that are all missing values. This may come in handy for a
         new line of business that doesn't have origins/developments of an
         existing line in the same triangle.
@@ -136,17 +210,44 @@ class TrianglePandas:
         min_odim = obj.origin[odim.index(1)]
         max_odim = obj.origin[::-1][odim[::-1].index(1)]
         if obj.shape[-1] != 1:
-            ddim = list((xp.nansum(obj.values[0, 0, :], -2) != 0).astype("int"))
+            ddim = list(
+                (xp.nansum(obj.values[0, 0, :], -2) != 0).astype("int"))
             ddim = obj.development[pd.Series(ddim).astype(bool)]
             obj = self[
-                (self.development >= ddim.min()) & (self.development <= ddim.max())
+                (self.development >= ddim.min()) & (
+                    self.development <= ddim.max())
             ]
             return obj[(self.origin >= min_odim) & (self.origin <= max_odim)]
         obj = self[(self.origin >= min_odim) & (self.origin <= max_odim)]
         return obj
 
+    def fillna(self, value=None, inplace=False):
+        """Fill nan with 'value' by axis.
+        Parameters
+        ----------
+        value: single value or array-like values, default = None
+            Value(s) to fill across the axis.
+
+        inplace: boolean, default = False
+            Whether to modify the triangle object directly (True), or
+            return a new modified triangle (False).
+
+        Returns
+        -------
+        Triangle
+        """
+        if inplace:
+            frame = self + value * 0
+            xp = self.get_array_module()
+            fill = (xp.nan_to_num(frame.values) == 0) * (self * 0 + value)
+            self.values = (frame + fill).values
+            return self
+        else:
+            new_obj = self.copy()
+            return new_obj.fillna(value=value, inplace=True)
+
     def drop(self, labels=None, axis=1):
-        """ Drop specified labels from rows or columns.
+        """Drop specified labels from rows or columns.
 
         Remove rows or columns by specifying label names and corresponding axis,
         or by specifying directly index or column names.
@@ -166,6 +267,7 @@ class TrianglePandas:
         Triangle
 
         """
+        labels = [labels] if type(labels) is str else list(labels)
         if axis == 1:
             return self[[item for item in self.columns if item not in labels]]
         else:
@@ -173,10 +275,10 @@ class TrianglePandas:
 
     @property
     def T(self):
-        return self.to_frame().T
+        return self.to_frame(origin_as_datetime=False).T
 
     def groupby(self, by, axis=0, *args, **kwargs):
-        """ Group Triangle by index values.  If the triangle is convertable to a
+        """Group Triangle by index values.  If the triangle is convertable to a
         DataFrame, then it defaults to pandas groupby functionality.
 
         Parameters
@@ -191,7 +293,7 @@ class TrianglePandas:
         return TriangleGroupBy(self, by, axis)
 
     def append(self, other):
-        """ Append rows of other to the end of caller, returning a new object.
+        """Append rows of other to the end of caller, returning a new object.
 
         Parameters
         ----------
@@ -207,7 +309,7 @@ class TrianglePandas:
         return concat((self, other), 0)
 
     def rename(self, axis, value):
-        """ Alter axes labels.
+        """Alter axes labels.
 
         Parameters
         ----------
@@ -235,7 +337,7 @@ class TrianglePandas:
         return self
 
     def astype(self, dtype, inplace=True):
-        """ Copy of the array, cast to a specified type.
+        """Copy of the array, cast to a specified type.
 
         Parameters
         ----------
@@ -248,7 +350,7 @@ class TrianglePandas:
         -------
         Triangle as new datatype
         """
-        obj = self.copy() if inplace is True else self
+        obj = self.copy() if inplace is False else self
         obj.values = obj.values.astype(dtype)
         return obj
 
@@ -262,67 +364,89 @@ class TrianglePandas:
         return self.iloc[self.index.sort_values(self.key_labels, *args, **kwargs).index]
 
     def exp(self):
-        xp = self.get_array_module()
-        obj = self.copy()
-        obj.values = xp.exp(obj.values)
-        return obj
+        return self.get_array_module().exp(self)
 
     def log(self):
-        xp = self.get_array_module()
-        obj = self.copy()
-        obj.values = xp.log(obj.values)
-        return obj
+        return self.get_array_module().log(self)
 
     def minimum(self, other):
-        obj = self.copy()
-        xp = self.get_array_module()
-        obj.values = xp.minimum(self.values, other)
-        return obj
+        return self.get_array_module().minimum(self, other)
 
     def maximum(self, other):
-        obj = self.copy()
-        xp = self.get_array_module()
-        obj.values = xp.maximum(self.values, other)
-        return obj
+        return self.get_array_module().maximum(self, other)
 
     def sqrt(self):
-        obj = self.copy()
-        xp = self.get_array_module()
-        obj.values = xp.sqrt(self.values)
-        return obj
+        return self.get_array_module().sqrt(self)
+
+    def round(self, decimals=0, *args, **kwargs):
+        return round(self, decimals)
 
 
-def add_triangle_agg_func(cls, k, v):
-    """ Aggregate Overrides in Triangle """
+def add_triangle_agg_func(
+        cls: Type[TrianglePandas],
+        k: str,
+        v: str
+):
+    """
+    Aggregate Overrides in Triangle
+    """
 
-    def agg_func(self, axis=None, *args, **kwargs):
+    def agg_func(
+            self: Triangle,
+            axis: str | int | None = None,
+            *args,
+            **kwargs
+    ) -> Triangle | np.ndarray:
+        """
+        Applies the aggregation function specified by k from the outer function.
+
+        Parameters
+        ----------
+
+        self: Triangle
+            The triangle to which the aggregation function will be applied.
+
+        axis: str | int | None
+            The axis of the triangle to which the aggregation function will be applied.
+
+        Returns
+        -------
+
+        An aggregated Triangle or ndarray.
+        """
         keepdims = kwargs.get("keepdims", None)
         obj = self.copy()
-        auto_sparse = kwargs.pop("auto_sparse", True)
+        auto_sparse: bool = kwargs.pop("auto_sparse", True)
         if axis is None:
             if max(obj.shape) == 1:
-                axis = 0
+                axis: int = 0
             else:
-                axis = min([num for num, _ in enumerate(obj.shape) if _ != 1])
+                axis: int = min([num for num, _ in enumerate(obj.shape) if _ != 1])
         else:
             axis = self._get_axis(axis)
-        xp = obj.get_array_module()
-        func = getattr(xp, v)
+        xp: ModuleType = obj.get_array_module()
+        func: Callable = getattr(xp, v)
         kwargs.update({"keepdims": True})
+        # Apply the function on the requested axis.
         obj.values = func(obj.values, axis=axis, *args, **kwargs)
+
+        # Aggregation function will collapse a dimension, so
+        # adjust the dimensions of the original object to match that of the aggregation.
         if axis == 0 and obj.values.shape[axis] == 1 and len(obj.kdims) > 1:
             obj.kdims = np.array([["(All)"] * len(obj.key_labels)])
         if axis == 1 and obj.values.shape[axis] == 1 and len(obj.vdims) > 1:
             obj.vdims = np.array([0])
         if axis == 2 and obj.values.shape[axis] == 1 and len(obj.odims) > 1:
             obj.odims = obj.odims[0:1]
+        # If axis is development, set the ddims to be the valuation date.
         if axis == 3 and obj.values.shape[axis] == 1 and len(obj.ddims) > 1:
             obj.ddims = pd.DatetimeIndex(
                 [self.valuation_date], dtype="datetime64[ns]", freq=None
             )
+        obj._set_slicers()
         if auto_sparse:
-            obj._set_slicers()
-        obj.values = num_to_nan(obj.values)
+            obj = obj._auto_sparse()
+        obj.values: ArrayLike = num_to_nan(obj.values)
         if not keepdims and obj.shape == (1, 1, 1, 1):
             return obj.values[0, 0, 0, 0]
         else:
@@ -331,28 +455,40 @@ def add_triangle_agg_func(cls, k, v):
     set_method(cls, agg_func, k)
 
 
-def add_groupby_agg_func(cls, k, v):
-    """ Aggregate Overrides in GroupBy """
+def add_groupby_agg_func(cls, k: str, v: str):
+    """Aggregate Overrides in GroupBy"""
 
     def agg_func(self, *args, **kwargs):
         from chainladder.utils import concat
-        from chainladder.methods import Chainladder
-
         xp = self.obj.get_array_module()
         obj = self.obj.copy()
-        values = [
-            getattr(obj.iloc.__getitem__(tuple([slice(None)] * self.axis + [i])), v)(
-                self.axis, auto_sparse=False, keepdims=True
-            )
-            for i in self.groups.indices.values()
-        ]
+        auto_sparse = kwargs.pop("auto_sparse", True)
+        if db and obj.array_backend == "sparse":
+
+            def aggregate(i, obj, axis, v):
+                return getattr(
+                    obj.iloc.__getitem__(tuple([slice(None)] * axis + [i])), v
+                )(axis, auto_sparse=False, keepdims=True)
+
+            bag = db.from_sequence(self.groups.indices.values())
+            bag = bag.map(aggregate, obj, self.axis, v)
+            values = bag.compute(scheduler="threads")
+        else:
+            values = [
+                getattr(
+                    obj.iloc.__getitem__(
+                        tuple([slice(None)] * self.axis + [i])), v
+                )(self.axis, auto_sparse=False, keepdims=True)
+                for i in self.groups.indices.values()
+            ]
         obj = concat(values, axis=self.axis, ignore_index=True)
+        group_index = self.groups.first().index
         if self.axis == 0:
-            if isinstance(self.groups.dtypes.index, pd.MultiIndex):
+            if isinstance(group_index, pd.MultiIndex):
                 index = (
                     pd.DataFrame(
-                        np.zeros(len(self.groups.dtypes.index)),
-                        index=self.groups.dtypes.index,
+                        np.zeros(len(group_index)),
+                        index=group_index,
                         columns=["_"],
                     )
                     .reset_index()
@@ -360,31 +496,33 @@ def add_groupby_agg_func(cls, k, v):
                 )
                 obj.index = index
             else:
-                index = pd.DataFrame(self.groups.dtypes.index)
+                index = pd.DataFrame(group_index)
                 obj.key_labels = index.columns.tolist()
                 obj.kdims = index.values
-        else:
-            index = pd.DataFrame(self.groups.dtypes.index).values[:, 0]
         if self.axis == 1:
-            obj.vdims = index
+            obj.vdims = pd.DataFrame(group_index).values[:, 0]
+        if self.axis == 2:
+            odims = self.obj._to_datetime(
+                pd.Series(self.groups.indices.keys()).to_frame(), [0]
+            )
+            obj.origin_grain = self.obj._get_grain(odims)
+            split = obj.origin_grain.split("-")
+            obj.origin_grain = {"A": "Y", "2Q": "S"}.get(split[0], split[0])
+            obj.odims = odims.values
         obj._set_slicers()
-        if hasattr(obj, 'ldf_'):
-            if len(obj.ldf_) > 1: # Bypass grouped ldf_ if there is only one anyway
-                new_ldf = Chainladder().fit(self.obj).full_expectation_
-                new_ldf = new_ldf.groupby(self.by).sum()  # Need to generalize sum
-                new_ldf = new_ldf.link_ratio.iloc[..., :self.obj.ldf_.shape[-1]]
-                if new_ldf.get_array_module().all(
-                    (new_ldf.values.max(2) - new_ldf.values.min(2)) < 1e-6):
-                    # if after grouping there is still only one, then compress to 1
-                    new_ldf = new_ldf.iloc[..., 0, :]
-                obj.ldf_ = new_ldf
+        if auto_sparse:
+            obj = obj._auto_sparse()
         return obj
 
-    set_method(cls, agg_func, k)
+    set_method(
+        cls=cls,
+        func=agg_func,
+        k=k
+    )
 
 
 def add_df_passthru(cls, k):
-    """Pass Through of pandas functionality """
+    """Pass Through of pandas functionality"""
 
     def df_passthru(self, *args, **kwargs):
         return getattr(pd.DataFrame, k)(self.to_frame(), *args, **kwargs)
@@ -392,17 +530,55 @@ def add_df_passthru(cls, k):
     set_method(cls, df_passthru, k)
 
 
-def set_method(cls, func, k):
-    """ Assigns methods to a class """
+def set_method(
+        cls: Type[TrianglePandas | TriangleGroupBy],
+        func: Callable,
+        k: str
+) -> None:
+    """
+    Assigns methods to a class.
+
+    Parameters
+    ----------
+
+    cls: Type[TrianglePandas | TriangleGroupBy]
+        Class to be modified.
+
+    func: Callable
+        Method to be added to the class supplied to parameter cls.
+
+    k: str
+        Name of the method to be added to the class supplied to parameter cls.
+
+    Returns
+    -------
+
+    None
+    """
     func.__doc__ = "Refer to pandas for ``{}`` functionality.".format(k)
     func.__name__ = k
     setattr(cls, func.__name__, func)
 
 
 df_passthru = (
-    ["to_clipboard", "to_csv", "to_excel", "to_json", "to_html",]
-    + ["to_dict", "unstack", "pivot", "drop_duplicates", "describe", "melt",]
-    + ["pct_chg", "round",]
+    [
+        "to_clipboard",
+        "to_csv",
+        "to_excel",
+        "to_json",
+        "to_html",
+    ]
+    + [
+        "to_dict",
+        "unstack",
+        "pivot",
+        "drop_duplicates",
+        "describe",
+        "melt",
+    ]
+    + [
+        "pct_chg",
+    ]
 )
 for item in df_passthru:
     add_df_passthru(TrianglePandas, item)
