@@ -21,54 +21,34 @@ class MethodBase(BaseEstimator, EstimatorIO, Common):
             obj = Development().fit_transform(obj)
         if len(obj.ddims) - len(obj.ldf_.ddims) == 1:
             obj = TailConstant().fit_transform(obj)
-        return obj
+        return obj.val_to_dev()
 
-    def _align_cdf(self, ultimate, sample_weight=None):
-        """ Vertically align CDF to ultimate vector to origin period latest
-        diagonal.
-        """
-        xp = ultimate.get_array_module()
-        from chainladder.utils.utility_functions import num_to_nan
-        if len(self.ldf_) != len(ultimate) and len(self.ldf_.index) > 1:
-            if hasattr(ultimate, 'group_index'):
-                group_index = ultimate.group_index
-            else:
-                group_index = ultimate.index
-            level = list(
-                set(group_index.columns).intersection(
-                set(self.ldf_.key_labels)))
-            idx = group_index.merge(
-                self.ldf_.index.reset_index(),
-                how='left', on=level)['index'].values.astype(int)
-            cdf = self.cdf_.values[list(idx), ..., : ultimate.shape[-1]]
-        else:
-            cdf = self.cdf_.values[..., : ultimate.shape[-1]]
-        a = ultimate.iloc[0, 0] * 0
+    def _align_cdf(self, X, sample_weight=None):
+        """ Vertically align CDF to origin period latest diagonal. """
+        valuation = X.valuation_date
+        cdf = X.cdf_.iloc[..., : X.shape[-1]]
+        a = X.iloc[0, 0] * 0
         a = a + a.nan_triangle
-        if ultimate.array_backend == "sparse":
+        if X.array_backend == "sparse":
             a = a - a[a.valuation < a.valuation_date]
-        a = a.set_backend(ultimate.array_backend)
         if sample_weight:
-            ultimate.values = xp.nan_to_num(ultimate.values * a.values) + xp.nan_to_num(
-                sample_weight.values * a.values
-            )
+            X = X * a + sample_weight * a
         else:
-            ultimate.values = xp.nan_to_num(ultimate.values * a.values)
-        ultimate.values = num_to_nan(ultimate.values)
-        ultimate = ultimate / ultimate
-        cdf = ultimate * cdf
-        cdf = cdf.latest_diagonal.values
-        return cdf
+            X = X * a
+        cdf = X / X * cdf
+        cdf.valuation_date = valuation
+        return cdf.latest_diagonal
 
     def _set_ult_attr(self, ultimate):
         """ Ultimate scaffolding """
-        from chainladder import ULT_VAL
+        from chainladder import options
 
         xp = ultimate.get_array_module()
         if ultimate.array_backend != "sparse":
             ultimate.values[~xp.isfinite(ultimate.values)] = xp.nan
-        ultimate.ddims = pd.DatetimeIndex([ULT_VAL])
+        ultimate.ddims = pd.DatetimeIndex([options.ULT_VAL])
         ultimate.virtual_columns.columns = {}
+        ultimate.is_cumulative = True
         ultimate._set_slicers()
         ultimate.valuation_date = ultimate.valuation.max()
         ultimate._drop_subtriangles()
@@ -80,7 +60,10 @@ class MethodBase(BaseEstimator, EstimatorIO, Common):
 
     @property
     def latest_diagonal(self):
-        return self.X_.latest_diagonal
+        if self.X_.is_cumulative:
+            return self.X_.latest_diagonal
+        else:
+            return self.X_.sum('development')
 
     def fit(self, X, y=None, sample_weight=None):
         """Applies the chainladder technique to triangle **X**
@@ -123,14 +106,35 @@ class MethodBase(BaseEstimator, EstimatorIO, Common):
         X_new: Triangle
 
         """
-        obj = X.copy()
-        xp = obj.get_array_module()
-        obj.ldf_ = self.ldf_
-        self.validate_weight(X, sample_weight)
+        X_new = X.val_to_dev()
+        if sum(X_new.ddims > self.ldf_.ddims.max()) > 0:
+            raise ValueError("X has ages that exceed those available in model.")
+        X_new = X_new + (self.X_.val_to_dev().iloc[0,0].sum(2) * 0)
+        self.validate_weight(X_new, sample_weight)
         if sample_weight:
-            sample_weight = sample_weight.set_backend(obj.array_backend)
-        obj.ultimate_ = self._get_ultimate(obj, sample_weight)
-        return obj
+            sample_weight = sample_weight.set_backend(X_new.array_backend)
+        X_new.ldf_ = self.ldf_
+        X_new, X_new.ldf_ = self.intersection(X_new, X_new.ldf_)
+        return X_new
+        
+    def intersection(self, a, b):
+        """ Given two Triangles with mismatched indices, this method aligns
+            their indices """
+        if len(a) == 1 and len(b) == 1:
+            return a, b
+        intersection = list(set(a.key_labels).intersection(set(b.key_labels)))
+        if intersection == []:
+            return a, b
+        a_idx = a.index[intersection]
+        b_idx = b.index[intersection]
+        idx_intersection = list(
+            set(a_idx.set_index(intersection).index.intersection(
+                b_idx.set_index(intersection).index)))
+        if (len(a) == 1 or len(b) == 1) and idx_intersection == []:
+            return a, b
+        b = b.iloc[b_idx[b_idx[intersection].set_index(intersection).index.isin(idx_intersection)].index]
+        a = a.iloc[a_idx[a_idx[intersection].set_index(intersection).index.isin(idx_intersection)].index]
+        return a, b
 
     def fit_predict(self, X, y=None, sample_weight=None):
         self.fit(X, y, sample_weight)
@@ -154,5 +158,4 @@ class MethodBase(BaseEstimator, EstimatorIO, Common):
             and sample_weight.shape[0] > 1
         ):
             warnings.warn(
-                "X and sample_weight are not aligned.  Broadcasting may occur.\n"
-            )
+                "X and sample_weight are not aligned. Broadcasting may occur.\n")
