@@ -6,7 +6,7 @@ from chainladder.methods import MethodBase
 
 class Chainladder(MethodBase):
     """
-    The basic determinsitic chainladder method.
+    The basic deterministic chainladder method.
 
     Parameters
     ----------
@@ -14,16 +14,16 @@ class Chainladder(MethodBase):
 
     Attributes
     ----------
-    X_
+    X_:
         returns **X** used to fit the triangle
-    ultimate_
+    ultimate_:
         The ultimate losses per the method
-    ibnr_
+    ibnr_:
         The IBNR per the method
-    full_expectation_
+    full_expectation_:
         The ultimates back-filled to each development period in **X** replacing
         the known data
-    full_triangle_
+    full_triangle_:
         The ultimates back-filled to each development period in **X** retaining
         the known data
     """
@@ -33,14 +33,14 @@ class Chainladder(MethodBase):
 
         Parameters
         ----------
-        X : Triangle-like
+        X: Triangle-like
             Data to which the model will be applied.
-        y : Ignored
+        y: Ignored
         sample_weight : Ignored
 
         Returns
         -------
-        self : object
+        self: object
             Returns the instance itself.
         """
         super().fit(X, y, sample_weight)
@@ -48,10 +48,29 @@ class Chainladder(MethodBase):
         self.process_variance_ = self._include_process_variance()
         return self
 
+    def predict(self, X, sample_weight=None):
+        """Predicts the chainladder ultimate on a new triangle **X**
+
+        Parameters
+        ----------
+        X: Triangle
+            Loss data to which the model will be applied.
+        sample_weight: Triangle
+            Required exposure to be used in the calculation.
+
+        Returns
+        -------
+        X_new: Triangle
+            Loss data with chainladder ultimate applied
+        """
+        X_new = super().predict(X, sample_weight)
+        X_new.ultimate_ = self._get_ultimate(X_new, sample_weight)
+        return X_new
+
     def _get_ultimate(self, X, sample_weight=None):
         """ Private method that uses CDFs to obtain an ultimate vector """
-        xp = X.get_array_module()
-        ultimate = X.copy()
-        cdf = self._align_cdf(ultimate, sample_weight)
-        ultimate = X.latest_diagonal * cdf
+        ld = X.incr_to_cum().latest_diagonal
+        ultimate = X.incr_to_cum().copy()
+        cdf = self._align_cdf(ultimate, sample_weight) 
+        ultimate = ld * cdf 
         return self._set_ult_attr(ultimate)
