@@ -11,22 +11,19 @@ class TailBase(DevelopmentBase):
         to development objects with an additional set of tail statistics"""
 
     def fit(self, X, y=None, sample_weight=None):
-        if X.array_backend == "sparse":
-            obj = X.set_backend("numpy")
-            xp = np
-        else:
-            xp = X.get_array_module()
-            obj = X.copy()
+        obj = X.copy()
         if "ldf_" not in obj:
             obj = Development().fit_transform(obj)
-        self._ave_period = {"Y": (1, 12), "Q": (4, 3), "M": (12, 1)}[
+        xp = obj.ldf_.get_array_module()
+        m = int(self.projection_period / 12)
+        self._ave_period = {"Y": (1 * m, 12), "Q": (4 * m, 3), "M": (12 * m, 1), "S": (2 * m, 6)}[
             obj.development_grain
         ]
         t_ddims = [
             (item + 1) * self._ave_period[1] + obj.ldf_.ddims[-1]
-            for item in range(self._ave_period[0])
+            for item in range(self._ave_period[0]+1)
         ]
-        ddims = np.concatenate((obj.ldf_.ddims, t_ddims, np.array([9999])), 0,)
+        ddims = np.concatenate((obj.ldf_.ddims, t_ddims), 0,)
         self.ldf_ = obj.ldf_.copy()
         tail = xp.ones(self.ldf_.shape)[..., -1:]
         tail = xp.repeat(tail, self._ave_period[0] + 1, -1)
@@ -64,7 +61,8 @@ class TailBase(DevelopmentBase):
         X_new = X.copy()
         triangles = ["ldf_", "std_err_", "sigma_"]
         for item in triangles + ["tail_", "_ave_period", "average_"]:
-            setattr(X_new, item, getattr(self, item))
+            if hasattr(self, item):
+                setattr(X_new, item, getattr(self, item))
         X_new._set_slicers()
         return X_new
 
@@ -124,7 +122,10 @@ class TailBase(DevelopmentBase):
             y = num_to_nan(y)
             reg = WeightedRegression(axis=3, xp=xp).fit(None, xp.log(y), None)
             std_err_ = xp.exp(time_pd * reg.slope_ + reg.intercept_)
-
+            if self.tail_.values.flatten().sum() / xp.prod(self.tail_.shape) == 1.0:
+                # If no tail, assume no variation
+                sigma_ = sigma_ * 0
+                std_err_ = std_err_* 0
             self.sigma_.values = xp.concatenate(
                 (self.sigma_.values[..., :-1], sigma_[..., -1:]), axis=-1
             )
@@ -157,7 +158,7 @@ class TailBase(DevelopmentBase):
             == self.cdf_.development.iloc[-1 - self._ave_period[0]]
         ]
         if np.all(df.values.min(axis=2) == df.values.max(axis=2)):
-            df = df.iloc[..., 0, :].to_frame()
+            df = df.iloc[..., 0, :].to_frame(origin_as_datetime = False)
         return df
 
     @property
